html {
    height: 100%;
}

body {
    width: 90%;
    height: 90%;
    margin: 0 auto;
    color: #333;
    font-weight: 300;
    font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
}

#main_chart {
    width: 100%;
    height: 90%;
    border: 1px solid #ccc;
    overflow: hidden;
    margin: 0 auto;
}

#diff_modal .modal-dialog {
    max-width: 800px
}

